/* fraid.c -- driver to use gdb as a binary file editor
  Copyright 1999, 2002 BitWagon Software LLC.  All rights reserved.
  This file may be copied, modified, and distributed under the terms
  of the GNU General Public License (GPL) version 2.
*/
#include <errno.h>
#include <stdio.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/mman.h>

int
main(int argc, char const *const *const argv, char const *const *const envp)
{
	struct stat st;
	int fd;
	int err1;
	int res2;
	int err2;
	int prot = PROT_EXEC | PROT_READ | PROT_WRITE;
	int flags = MAP_FIXED | MAP_SHARED;
	char *addr;

	if (2!=argc) {
		fprintf(stderr, "Usage: gdb fraid; (gdb) run <file>\n");
		exit(1);
	}
	fd = open(argv[1], O_RDWR, 0);
	err1 = errno;
	if (fd < 0) {
		perror(argv[1]);
		fprintf(stderr, "Trying readonly...\n");
		flags = MAP_FIXED | MAP_PRIVATE;
		fd = open(argv[1], O_RDONLY, 0);
	}
	res2 = fstat(fd, &st);
	err2 = errno;

	if (fd < 0) {
		errno = err1;
		perror(argv[1]);
		exit(1);
	}
	if (0!=res2) {
		errno = err2;
		perror(argv[1]);
		exit(1);
	}
	addr = mmap(0, st.st_size, prot, flags, fd, 0);
	if (0!=addr) {
		perror(argv[1]);
		exit(1);
	}
	fprintf(stderr,"%d (%#x) bytes\n", st.st_size, st.st_size);
	__asm__("int3");
	return 0;
}

