/* tub_cstart.c -- take control of virtual address space from TASK_UNMAPPED_BASE.
   Copyright 2003 BitWagon Software LLC.  All rights reserved.
   This file may be copied, modified, and distributed under the terms
   of the GNU General Public License (GPL) version 2.
*/
#include <sys/types.h>
#include <sys/user.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <elf.h>

static void
__attribute__ ((regparm(2), stdcall))
bzero(char *p, size_t len)
{
    if (len) do {
        *p++= 0;
    } while (--len);
}

extern void *task_unmapped_base;
extern void *  tub_mmap(void *addr, size_t len, int prot, int flags, int fd, off_t offset);
extern void *tub_munmap(void *addr, size_t len);

/*#include <asm/unistd.h>*/
extern void     claim_pages(void const *const addr, size_t const len, int const need_lock);
extern unsigned alloc_pages(void const *const addr, size_t const len, void const *const base);
extern void      free_pages(void const *const addr, size_t const len);

#define REP8(x) \
    ((x)|((x)<<4)|((x)<<8)|((x)<<12)|((x)<<16)|((x)<<20)|((x)<<24)|((x)<<28))
#define EXP8(y) \
    ((1&(y)) ? 0xf0f0f0f0 : (2&(y)) ? 0xff00ff00 : (4&(y)) ? 0xffff0000 : 0)

static void *
xfind_pages(Elf32_Phdr const *phdr, int phnum)
{
	size_t lo= ~0, hi= 0;
	for (; --phnum>=0; ++phdr) if (PT_LOAD==phdr->p_type) {
		if (phdr->p_vaddr < lo) {
			lo = phdr->p_vaddr;
		}
		if (hi < (phdr->p_memsz + phdr->p_vaddr)) {
			hi =  phdr->p_memsz + phdr->p_vaddr;
		}
	}
	lo &= PAGE_MASK;
	hi = PAGE_MASK & (~PAGE_MASK + hi);
    hi-= lo;  /* length */

    lo = alloc_pages((void *)lo, hi, task_unmapped_base);
    lo<<= PAGE_SHIFT;
    free_pages((void *)lo, hi);
    return (void *)lo;
}

static unsigned long
do_xmap(int fdi, Elf32_Ehdr *const ehdr)
{
	unsigned first_vaddr = 1;
	Elf32_Phdr const *phdr = (Elf32_Phdr const *) (ehdr->e_phoff +
		(char const *)ehdr);
	unsigned long base = (unsigned long)xfind_pages(phdr, ehdr->e_phnum);
	int j;

	if (fdi < 0) {
		goto fail;
	}
	/* base is const from now on */
	for (j=0; j < ehdr->e_phnum; ++phdr, ++j)
	if (PT_LOAD==phdr->p_type && phdr->p_memsz!=0) {
		int const prot = 7 & (( (REP8(PROT_EXEC ) & EXP8(PF_X))
                               |(REP8(PROT_READ ) & EXP8(PF_R))
                               |(REP8(PROT_WRITE) & EXP8(PF_W)) )
                             >> ((phdr->p_flags & (PF_R|PF_W|PF_X))<<2) );
		size_t frag = ~PAGE_MASK & phdr->p_vaddr;  /* at lo end */
		size_t mlen = frag + phdr->p_filesz;
		char *addr, *got;

		if (1==first_vaddr) {
			first_vaddr = PAGE_MASK & phdr->p_vaddr;
		}
		addr = (char *)(phdr->p_vaddr - frag - first_vaddr + base);
		got = tub_mmap(addr, mlen, prot, MAP_FIXED | MAP_PRIVATE,
				fdi, phdr->p_offset - frag);
		if (got!=addr) {
fail:
			err_exit(2);
		}
		if (0!=frag && phdr->p_flags & PF_W) {
			bzero(addr, frag);  /* fragment at lo end */
		}
		addr += mlen;
		frag = (-mlen) &~ PAGE_MASK;  /* distance to next page boundary */
		if (0!=frag && phdr->p_flags & PF_W) {
			bzero(addr, frag);  /* fragment at hi end */
		}
		if (phdr->p_memsz != phdr->p_filesz) { /* .bss */
			addr += frag;
			if ((addr - got) < phdr->p_memsz) { /* need more pages, too */
				mlen = phdr->p_memsz - (addr - got);
				if (addr != tub_mmap(addr, mlen, prot,
						MAP_FIXED | MAP_PRIVATE | MAP_ANONYMOUS, 0, 0 ) ) {
					goto fail;
				}
			}
		}
	}
	if (0!=tub_close(fdi)) {
		goto fail;
	}
	return ehdr->e_entry - first_vaddr + base;
}

Elf32_Addr
__attribute__ ((regparm(1), stdcall))
tub_cstart(unsigned const *const esp)
{
    Elf32_Phdr *phdr = 0;
    int phnum = 0;
    Elf32_auxv_t *av;
    int const argc = *esp;
    {
        unsigned const *p = 2+ argc + esp;
        while (0!=*p++) ;
        av = (Elf32_auxv_t *)p;
    }
    {
        unsigned needed = (1<<AT_NULL) | (1<<AT_PHDR) | (1<<AT_PHNUM) | (1<<AT_ENTRY);
        for (; AT_NULL!=av->a_type; ++av) {
            if (AT_PHDR==av->a_type) {
                phdr = (Elf32_Phdr *)av->a_un.a_ptr;
                needed &= ~(1<<AT_PHDR);
            }
            if (AT_PHNUM==av->a_type) {
                phnum = av->a_un.a_val;
                needed &= ~(1<<AT_PHNUM);
            }
            if (AT_ENTRY==av->a_type) {
                extern void _start(void);
                av->a_un.a_ptr = _start;
                needed &= ~(1<<AT_ENTRY);
            }
        }
        for (; --phnum>=0; ++phdr) if (PT_NOTE==phdr->p_type) {
            tub_iprotect(phdr, sizeof(*phdr), PROT_WRITE | PROT_READ);
            phdr->p_type = PT_INTERP;
            tub_iprotect(phdr, sizeof(*phdr), PROT_EXEC  | PROT_READ);
            needed &= ~(1<<AT_NULL);
            break;
        } else if (PT_LOAD==phdr->p_type) {
            claim_pages((void const *)phdr->p_vaddr, phdr->p_memsz, 0);
        }
        if (0!=needed) {
            static char const msg[]="TUB: bad Elf.\n";
            tub_write(2, msg, strlen(msg));
            err_exit(1);
        }
    }
    {
        int const fd = tub_open((char const *)phdr->p_vaddr, O_RDONLY);
        unsigned char buf[512];
        task_unmapped_base = (void *)(PAGE_MASK & (~PAGE_MASK + tub_brk(0)));
        task_unmapped_base += 10*-PAGE_MASK;  /* allow sbrk() to work a little */

        if (sizeof(buf)!=tub_read(fd, buf, sizeof(buf))) {
            err_exit(11);
        }
        return do_xmap(fd, (Elf32_Ehdr *)buf);
    }
}

/*
vi:ts=4:et:
*/
