/* tub_libc.S -- syscalls in the wilderness; assembly language glue for mmap.
   Copyright 2003 BitWagon Software LLC.  All rights reserved.
   This file may be copied, modified, and distributed under the terms
   of the GNU General Public License (GPL) version 2.
*/
#include <asm/unistd.h>

	.globl tub_start
tub_start:
	nop  # int3
	movl %esp,%eax; call tub_cstart
	jmp *%eax
	
tub_unlock: .globl tub_unlock  # tub_unlock(&word)
	movl 4(%esp),%ecx
	subl %eax,%eax
	movl %eax,(%ecx)
	ret

tub_lock: .globl tub_lock  # tub_lock(&word)
	movl $0xfffff000,%ecx  # syscall error threshhold
	movl  $ __NR_gettid,      %eax; int $0x80; cmpl %ecx,%eax; jbe tl0
	pushl $ __NR_getpid; popl %eax; int $0x80; cmpl %ecx,%eax; jbe tl0
	pushl $1; popl %eax
tl0:
	movl 4(%esp),%ecx  # &lock
	movl %eax,%edx  # our identity for successful locking
tl1:
	subl %eax,%eax  # value when lock is open

	/* if (eax == *ecx) { *ecx = edx; } else { eax = *ecx; } */
	lock; cmpxchg %edx,(%ecx); jne tl2  # already locked

	ret  # success: 0 in %eax; lock has thread identity
tl2:
	pause  # keep Pentium4 cool
	cmpl %edx,%eax; jne tl1  # not this thread, therefore spin

	.pushsection ".rodata"
lockmsg: .ascii "TUB: double lock; signal handler?\n"
len_lockmsg = . - lockmsg
	.popsection

	pushl %ebx
	pushl $ len_lockmsg
	pushl $ lockmsg
	pushl $2  # FILENO(stderr)
	call tub_write; addl $3*4,%esp
	hlt

tub_mmap: .globl tub_mmap
	movl %ebx,%edx; leal 4(%esp),%ebx
	call l_pre_mmap
	int $0x80
	movl %edx,%ebx
	call l_post_mmap  # ja err_sys  too early [do_xmap() knows this]
	ret

tub_munmap: .globl tub_munmap
	          movl %ebx,%edx
	movl 1*4(%esp),%ebx
	movl 2*4(%esp),%ecx
	pushl $ __NR_munmap; popl %eax
	int $0x80
	          movl %edx,%ebx
	call l_post_munmap
	ret

__pre_mmap: .globl __pre_mmap
l_pre_mmap:
	pushl %ecx; pushl %edx
	call tub_pre_mmap
	popl %edx; popl %ecx
	ret

__post_mmap: .globl __post_mmap
l_post_mmap:
	pushl %ecx; pushl %edx
	call tub_post_mmap
	popl %edx; popl %ecx
	cmpl $0xfffff000,%eax  # replacement
	ret

__pre_mmap64: .globl __pre_mmap64
l_pre_mmap64:
	pushal
	call tub_pre_mmap64; movl %eax,7*4(%esp)
	popal
	ret

__post_mmap64: .globl __post_mmap64
l_post_mmap64:
	pushal
	call tub_post_mmap64; movl %eax,7*4(%esp)
	popal
	cmpl $0xfffff000,%eax  # replacement
	ret

__post_munmap: .globl __post_munmap
l_post_munmap:
	pushl %ecx; pushl %edx
	call tub_post_munmap
	popl %edx; popl %ecx
	cmpl $0xfffff000,%eax  # replacement
	ret

PAGE_MASK=~0<<12

done:
	ret
tub_iprotect:.globl tub_iprotect; .type tub_iprotect,@function  # interval
	movl 2*4(%esp),%ecx  # length
	movl 1*4(%esp),%eax  # addr
	testl %ecx,%ecx; je done
	movl $ PAGE_MASK,%edx
	leal -1(%eax,%ecx),%ecx  # &last affected
	andl %edx,%eax  # round down
	andl %edx,%ecx
	subl %eax,%ecx  # page-crossing length
	subl %edx,%ecx  # total length
	movl %eax,1*4(%esp)  # new addr
	movl %ecx,2*4(%esp)  # new length
tub_mprotect:.globl tub_mprotect; .type tub_mprotect,@function; .size tub_mprotect,4
	movb $ __NR_mprotect,%al; jmp sys3

tub_write: .globl tub_write; .type tub_write,@function; .size tub_write,4
	movb $ __NR_write,%al; jmp sys3
tub_read: .globl tub_read; .type tub_read,@function; .size tub_read,4
	movb $ __NR_read,%al; jmp sys3
tub_open: .globl tub_open; .type tub_open,@function; .size tub_open,4
	movb $ __NR_open,%al; jmp sys3
tub_close: .globl tub_close; .type tub_close,@function; .size tub_close,4
	movb $ __NR_close,%al; jmp sys3
tub_brk: .globl tub_brk; .type tub_brk,@function; .size tub_brk,4
	movb $ __NR_brk,%al; jmp sys3
sys3:
	pushl %edi; pushl %esi  # for benefit of merge with sys5
sys3go:
	pushl %ebx
	movzbl %al,%eax
	movl (3+1+2)*4(%esp),%edx
	movl (3+1+1)*4(%esp),%ecx
	movl (3+1+0)*4(%esp),%ebx
sysgo:
	int $0x80
	popl %ebx
	popl %esi; popl %edi
	cmpl $0xffffff00,%eax; ja err_sys
	ret

rt_sigaction: .globl rt_sigaction; .type rt_sigaction,@function; .size rt_sigaction,4  # sig, newact, oldact, sigsiz
	movb $ __NR_rt_sigaction,%al
	pushl %edi; pushl %esi
	movl (2+1+3)*4(%esp),%esi
	jmp sys3go

	.pushsection ".data"
tub_errno: .globl tub_errno
	.long 0
	.popsection

err_sys:
	negl %eax; movl %eax,tub_errno
	orl $~0,%eax
	ret

err_exit: .globl err_exit
	hlt
